"""
==================================
Plotting memory consumptions
==================================

This example compares memory used when using a ResNet or a Momentum ResNet as a function of their depth

"""  # noqa

# Authors: Michael Sander, Pierre Ablin
# License: MIT

import torch
import torch.nn as nn
from momentumnet import MomentumNet, Mom
import matplotlib.pyplot as plt
from memory_profiler import memory_usage
import numpy as np
import os

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


if not os.path.isdir("figures"):
    os.mkdir("figures")


###########################################
# Fix random seed for reproducible figures
###########################################

torch.manual_seed(1)

##############################
# Parameters of the simulation
##############################


Depths = np.arange(1, 300, 100)


hidden = 8
gamma = 0.99
d = 500

function = nn.Sequential(nn.Linear(d, hidden), nn.Tanh(), nn.Linear(hidden, d))
function_res = nn.Sequential(
    nn.Linear(d, hidden), nn.Tanh(), nn.Linear(hidden, d)
)

X = torch.rand(500, 500)


def train(net):
    Loss = (net(X) ** 2).mean()
    Loss.backward()


if __name__ == "__main__":
    Mem_list_mom = []

    for n_iters in Depths:
        mom_net = Mom(
            [
                function,
            ]
            * n_iters,
            gamma=1 - 1 / (50 * n_iters),
            init_speed=0,
        )
        used_mem = np.max(memory_usage((train, (mom_net,))))
        Mem_list_mom.append(used_mem)

    Mem_list_res = []

    for n_iters in Depths:
        res_net = MomentumNet(
            [
                function_res,
            ],
            gamma=0.0,
            n_iters=n_iters,
            learn_gamma=False,
            init_speed=0,
        )
        used_mem = np.max(memory_usage((train, (res_net,))))
        Mem_list_res.append(used_mem)

    plt.figure(figsize=(4, 1.5))

    plt.plot(
        Depths, Mem_list_res, label="ResNet", linewidth=4, color="darkblue"
    )
    plt.plot(
        Depths, Mem_list_mom, label="MomentumNet", linewidth=4, color="red"
    )
    plt.yscale("log")
    y_ = plt.ylabel("Memory (MiB)")
    x_ = plt.xlabel("Depth")
    plt.legend()
    plt.show()



"""
==================================
Plotting the dynamics in 1D
==================================

This example compares the dynamics of a ResNet and a Momentum ResNet

"""  # noqa

# Authors: Michael Sander, Pierre Ablin
# License: MIT

import torch
import torch.nn as nn
import numpy as np
import matplotlib.pyplot as plt
import torch.optim as optim
from momentumnet import MomentumNet
from momentumnet.toy_datasets import make_data_1D
import os

if not os.path.isdir("figures"):
    os.mkdir("figures")
###########################################
# Fix random seed for reproducible figures
###########################################

torch.manual_seed(1)

##############################
# Parameters of the simulation
##############################

hidden = 16
n_iters = 15
gamma = 0.99
d = 1
function = nn.Sequential(nn.Linear(d, hidden), nn.Tanh(), nn.Linear(hidden, d))
function_res = nn.Sequential(
    nn.Linear(d, hidden), nn.Tanh(), nn.Linear(hidden, d)
)

mom_net = MomentumNet(
    [
        function,
    ],
    gamma=gamma,
    n_iters=n_iters,
    learn_gamma=False,
    init_speed=0,
)
res_net = MomentumNet(
    [
        function_res,
    ],
    gamma=0.0,
    n_iters=n_iters,
    learn_gamma=False,
    init_speed=0,
)


def h(x):
    return -(x ** 3)


def Loss(pred, x):
    return ((pred - h(x)) ** 2).mean()


optimizer = optim.SGD(mom_net.parameters(), lr=0.01)

# Training

print("MomentumNet ->")
for i in range(301):
    optimizer.zero_grad()
    x = make_data_1D(200)
    pred = mom_net(x)
    loss = Loss(pred, x)
    loss.backward()
    optimizer.step()
    if i % 100 == 0:
        print("- " * 20)
        print("itr %s, loss = %.3f" % (i, loss.item()))

print("- " * 40)

optimizer = optim.SGD(res_net.parameters(), lr=0.01)


print("ResNet -->")
for i in range(2001):
    optimizer.zero_grad()
    x = make_data_1D(200)
    pred = res_net(x)
    loss = Loss(pred, x)
    loss.backward()
    optimizer.step()
    if i % 100 == 0:
        print("- " * 20)
        print("itr %s, loss = %.3f" % (i, loss.item()))


# Plot the output


n_plot = 8

num_plots = n_plot

plt.figure(figsize=(3, 4))
colormap = plt.cm.gist_ncar
plt.gca().set_prop_cycle(
    plt.cycler("color", plt.cm.jet(np.linspace(0.8, 0.95, num_plots)))
)

x_ = make_data_1D(n_plot)
x = np.linspace(-1, 1, n_plot)
x_ = torch.tensor(x).view(-1, d).float()
x_axis = np.arange(0, n_iters + 1)

preds = np.zeros((n_iters + 1, n_plot))

preds[0] = x_[:, 0]

for i in range(1, n_iters + 1):
    with torch.no_grad():
        pred_ = mom_net(x_, n_iters=i)
        preds[i] = pred_[:, 0]

plt.plot(preds, x_axis, "-x", lw=2.5)
plt.xticks([], [])
plt.yticks([], [])
plt.ylabel("Depth")
plt.xlabel("Input")
plt.savefig("figures/mom_net_dynamics_1D.pdf")

num_plots = n_plot

plt.figure(figsize=(3, 4))
colormap = plt.cm.gist_ncar
plt.gca().set_prop_cycle(
    plt.cycler("color", plt.cm.jet(np.linspace(0.0, 0.1, num_plots)))
)

x_axis = np.arange(0, n_iters + 1)

preds_res = np.zeros((n_iters + 1, n_plot))

preds_res[0] = x_[:, 0]
for i in range(1, n_iters + 1):
    with torch.no_grad():
        pred_ = res_net(x_, n_iters=i)
        preds_res[i] = pred_[:, 0]

plt.plot(preds_res, x_axis, "-x", lw=2.5)
plt.xticks([], [])
plt.yticks([], [])
plt.ylabel("Depth")
plt.xlabel("Input")
plt.savefig("figures/res_net_dynamics_1D.pdf")



# Authors: Michael Sander, Pierre Ablin
# License: MIT

import os
import datetime
import time
import warnings
import numpy as np
import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.optim
import torch.utils.data
import torch.utils.data.distributed
import torchvision.transforms as transforms
import torchvision.datasets as datasets

from .models import (
    ResNet18,
    mResNet18,
    ResNet34,
    mResNet34,
    ResNet101,
    mResNet101,
    ResNet152,
    mResNet152,
)


def main(
    datapath,
    saveadr,
    expe_name,
    arch,
    n_workers,
    n_epochs,
    batch_size,
    lr,
    momentum=0.9,
    weight_decay=1e-4,
    print_freq=100,
    init_speed=0,
):

    model_save_adr = "models/%s/%s_checkpoint.pth.tar" % (expe_name, saveadr)
    print("saving model at %s" % model_save_adr)
    resume = os.path.exists(model_save_adr)

    def train(train_loader, model, criterion, optimizer, epoch):
        batch_time = AverageMeter("Time", ":6.3f")
        data_time = AverageMeter("Data", ":6.3f")
        losses = AverageMeter("Loss", ":.4e")
        top1 = AverageMeter("Acc@1", ":6.2f")
        top5 = AverageMeter("Acc@5", ":6.2f")
        progress = ProgressMeter(
            len(train_loader),
            [batch_time, data_time, losses, top1, top5],
            prefix="Epoch: [{}]".format(epoch),
        )

        # switch to train mode
        model.train()

        end = time.time()
        for i, (images, target) in enumerate(train_loader):
            # measure data loading time
            data_time.update(time.time() - end)

            if gpu is not None:
                images = images.cuda(gpu, non_blocking=True)
            if torch.cuda.is_available():
                target = target.cuda(gpu, non_blocking=True)

            # compute output
            output = model(images)
            loss = criterion(output, target)

            # measure accuracy and record loss
            acc1, acc5 = accuracy(output, target, topk=(1, 5))
            losses.update(loss.item(), images.size(0))
            top1.update(acc1[0], images.size(0))
            top5.update(acc5[0], images.size(0))

            # compute gradient and do SGD step
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            if i % print_freq == 0:
                progress.display(i)
        return losses.avg, top1.avg.item(), top5.avg.item()

    def validate(val_loader, model, criterion):
        batch_time = AverageMeter("Time", ":6.3f")
        losses = AverageMeter("Loss", ":.4e")
        top1 = AverageMeter("Acc@1", ":6.2f")
        top5 = AverageMeter("Acc@5", ":6.2f")
        progress = ProgressMeter(
            len(val_loader), [batch_time, losses, top1, top5], prefix="Test: "
        )

        # switch to evaluate mode
        model.eval()

        with torch.no_grad():
            end = time.time()
            for i, (images, target) in enumerate(val_loader):
                if gpu is not None:
                    images = images.cuda(gpu, non_blocking=True)
                if torch.cuda.is_available():
                    target = target.cuda(gpu, non_blocking=True)

                # compute output
                output = model(images)
                loss = criterion(output, target)

                # measure accuracy and record loss
                acc1, acc5 = accuracy(output, target, topk=(1, 5))
                losses.update(loss.item(), images.size(0))
                top1.update(acc1[0], images.size(0))
                top5.update(acc5[0], images.size(0))

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                if i % print_freq == 0:
                    progress.display(i)

            print(
                " * Acc@1 {top1.avg:.3f} Acc@5 {top5.avg:.3f}".format(
                    top1=top1, top5=top5
                )
            )

        return losses.avg, top1.avg.item(), top5.avg.item()

    gpu = None
    if gpu is not None:
        warnings.warn(
            "You have chosen a specific GPU. This will completely "
            "disable data parallelism."
        )

    if gpu is not None:
        print("Use GPU: {} for training".format(gpu))
    # create model
    if arch == "resnet18":
        model = ResNet18(1000)
    if arch == "mresnet18":
        model = mResNet18(1000, init_speed=init_speed)
    if arch == "resnet34":
        model = ResNet34(1000)
    if arch == "mresnet34":
        model = mResNet34(1000, init_speed=init_speed)
    if arch == "resnet101":
        model = ResNet101(1000)
    if arch == "mresnet101":
        model = mResNet101(1000, init_speed=init_speed)
    if arch == "resnet152":
        model = ResNet152(1000)
    if arch == "mresnet152":
        model = mResNet152(1000, init_speed=init_speed)
    if not torch.cuda.is_available():
        print("using CPU, this will be slow")
    if gpu is not None:
        torch.cuda.set_device(gpu)
        model = model.cuda(gpu)
    else:
        model = torch.nn.DataParallel(model).cuda()

    # define loss function (criterion) and optimizer
    criterion = nn.CrossEntropyLoss().cuda(gpu)

    optimizer = torch.optim.SGD(
        model.parameters(), lr, momentum=momentum, weight_decay=weight_decay
    )

    # # optionally resume from a checkpoint
    start_epoch = 0
    train_losss, train_top1s, train_top5s = [], [], []
    test_losss, test_top1s, test_top5s = [], [], []
    if resume:
        print("=> loading checkpoint '{}'".format(model_save_adr))
        if gpu is None:
            checkpoint = torch.load(model_save_adr)
        else:
            # Map model to be loaded to specified single gpu.
            loc = "cuda:{}".format(gpu)
            checkpoint = torch.load(model_save_adr, map_location=loc)
        start_epoch = checkpoint["epoch"]
        print("start epoch : %d" % start_epoch)
        model.load_state_dict(checkpoint["state_dict"])
        optimizer.load_state_dict(checkpoint["optimizer"])
        train_losss = checkpoint["train_loss"]
        train_top1s = checkpoint["train_top1"]
        train_top5s = checkpoint["train_top5"]
        test_losss = checkpoint["test_loss"]
        test_top1s = checkpoint["test_top1"]
        test_top5s = checkpoint["test_top5"]

    cudnn.benchmark = True

    # Data loading code
    traindir = os.path.join(datapath, "train")
    valdir = os.path.join(datapath, "val")
    normalize = transforms.Normalize(
        mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
    )

    train_dataset = datasets.ImageFolder(
        traindir,
        transforms.Compose(
            [
                transforms.RandomResizedCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                normalize,
            ]
        ),
    )

    train_sampler = None

    train_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=batch_size,
        shuffle=(train_sampler is None),
        num_workers=n_workers,
        pin_memory=True,
        sampler=train_sampler,
    )

    val_loader = torch.utils.data.DataLoader(
        datasets.ImageFolder(
            valdir,
            transforms.Compose(
                [
                    transforms.Resize(256),
                    transforms.CenterCrop(224),
                    transforms.ToTensor(),
                    normalize,
                ]
            ),
        ),
        batch_size=batch_size,
        shuffle=False,
        num_workers=n_workers,
        pin_memory=True,
    )
    print("Start training")
    for epoch in range(start_epoch, n_epochs):
        t0 = time.time()
        adjust_learning_rate(optimizer, epoch, lr)

        # train for one epoch
        train_loss, train_top1, train_top5 = train(
            train_loader, model, criterion, optimizer, epoch
        )

        # evaluate on validation set
        test_loss, test_top1, test_top5 = validate(
            val_loader, model, criterion
        )
        train_losss.append(train_loss)
        train_top1s.append(train_top1)
        train_top5s.append(train_top5)
        test_losss.append(test_loss)
        test_top1s.append(test_top1)
        test_top5s.append(test_top5)
        np.save(
            "results/%s/%s.npy" % (expe_name, saveadr),
            np.array(
                [
                    train_losss,
                    train_top1s,
                    train_top5s,
                    test_losss,
                    test_top1s,
                    test_top5s,
                ]
            ),
        )
        print(
            "Epoch %d took %s"
            % (epoch, str(datetime.timedelta(seconds=int(time.time() - t0))))
        )

        save_checkpoint(
            {
                "epoch": epoch + 1,
                "arch": arch,
                "state_dict": model.state_dict(),
                "optimizer": optimizer.state_dict(),
                "train_loss": train_losss,
                "train_top1": train_top1s,
                "train_top5": train_top5s,
                "test_loss": test_losss,
                "test_top1": test_top1s,
                "test_top5": test_top5s,
            },
            filename=model_save_adr,
        )


def save_checkpoint(state, filename="checkpoint.pth.tar"):
    torch.save(state, filename)


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, name, fmt=":f"):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print("\t".join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = "{:" + str(num_digits) + "d}"
        return "[" + fmt + "/" + fmt.format(num_batches) + "]"


def adjust_learning_rate(optimizer, epoch, lr):
    """Sets the learning rate to
     the initial LR decayed by 10 every 30 epochs"""
    lr = lr * (0.1 ** (epoch // 5))
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top
    predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res



# Authors: Michael Sander, Pierre Ablin
# License: MIT
"""
Illustration of the drop-in replacement aspect of MomentumNets.

- Use example/from_resnet_to_momentumnet.py to transform ResNets
into MomentumNets, pretrained or not.

"""

from momentumnet.models import (
    BasicBlock,
    MBasicBlock,
    Bottleneck,
    MBottleneck,
    MResNet,
)


def transform(model, pretrained=False, gamma=0.9, mem=False):
    """Return the MomentumNet counterpart of the model


    Parameters
    ----------
    model : a torchvision model
        The resnet one desires to turn into a momentumnet
    pretrained : bool (default: False)
        Whether using a pretrained resnet and transfer its weights
    gamma : float (default: 0.9)
        The momentum term for the MomentumNet
    mem : bool (default: False)
        If True then the MomentumNet has a smaller memory footprint

    Return
    ------
    mresnet : the MomentumNet counterpart of model

    """
    resnet = model
    layers = [
        len(resnet.layer1),
        len(resnet.layer2),
        len(resnet.layer3),
        len(resnet.layer4),
    ]
    num_classes = resnet.fc.out_features
    try:
        _ = resnet.layer1[0].conv3
        mresnet = MResNet(
            Bottleneck, MBottleneck, layers, num_classes, gamma=gamma, mem=mem
        )
    except AttributeError:
        mresnet = MResNet(
            BasicBlock, MBasicBlock, layers, num_classes, gamma=gamma, mem=mem
        )
    params1 = resnet.named_parameters()
    params2 = mresnet.named_parameters()
    if pretrained:
        for (name1, param1), (name2, param2) in zip(params1, params2):
            param2.data.copy_(param1)
    return mresnet



# Authors: Michael Sander, Pierre Ablin
# License: MIT

"""Dataset utilities, for simulated examples
"""
import torch
import numpy as np


###########################################
# Fix random seed for reproducible figures
###########################################

torch.manual_seed(1)


def make_circle(radius=1, N=100, label=0, noise=0.2):
    """Create points clouds on a circle


    Parameters
    ----------
    radius : float (default: 1000)
        Radius of the circle
    N : int (default: 100)
        Number of points in the circle
    label : int (default: 0)
        Label of the points
    noise : float (default: .2)
        The noise

    Return
    ------
    x : tensor, shape (2, N)
        points
    y : tensor, shape (1, N)
        labels of the points

    """

    y = label * torch.ones(N)
    theta = 2 * np.pi * torch.rand(1, N)
    r = 3 * (radius + noise * torch.rand(1, N))
    x = torch.vstack((np.cos(theta) * r, np.sin(theta) * r))
    return x, y


def make_data(N=100, n_r=4):
    """Create the data to separate


    Parameters
    ----------

    N : int (default: 100)
        Number of points in each ring

    n_r : int (default: .2)
        Number of rings

    Return
    ------
    x : tensor, shape (N, 2)
        points
    y : tensor, shape (N, 1)
        labels of the points

    """

    radius = np.arange(1, n_r + 1) * 0.5
    y = torch.ones(n_r)
    y[::2] = 0
    data = [make_circle(r, N, label) for r, label in zip(radius, y)]
    x = np.concatenate([d[0] for d in data], axis=1).T
    y = np.concatenate([d[1] for d in data], axis=None)
    x = torch.tensor(x).float()
    y = torch.tensor(y).long()
    return x, y


def random_points_1D(N=100):
    """Sample random points in 1d


    Parameters
    ----------

    N : int (default: 100)
        Number of points

    Return
    ------
    x : tensor, shape (1, N)
        points

    """

    x = 2 * (torch.rand(1, N) - 0.5)
    return x


def make_data_1D(N):
    x = random_points_1D(N)
    return x.view((-1, 1))



# Authors: Michael Sander, Pierre Ablin
# License: MIT

import torch
import torch.nn as nn

from momentumnet import MomentumNet, Mom
from numpy.testing import assert_raises

###########################################
# Fix random seed for reproducible figures
###########################################

torch.manual_seed(1)


def test_dimension_layers():
    function = nn.Sequential(nn.Linear(3, 3), nn.Linear(3, 3), nn.Linear(3, 3))
    mom_net = MomentumNet(
        [
            function,
        ],
        gamma=0.9,
    )
    x = torch.rand(3)
    assert mom_net(x).shape == x.shape
    function = nn.Sequential(nn.Linear(3, 4), nn.Linear(3, 3), nn.Linear(3, 3))
    mom_net = MomentumNet(
        [
            function,
        ],
        gamma=0.9,
    )
    assert_raises(RuntimeError, mom_net, x)
    function = nn.Sequential(nn.Linear(3, 3), nn.Linear(3, 3), nn.Linear(3, 3))
    mom = Mom(
        [
            function,
        ],
        gamma=0.9,
    )
    assert mom(x).shape == x.shape
    function = nn.Sequential(nn.Linear(3, 4), nn.Linear(3, 3), nn.Linear(3, 3))
    mom = Mom(
        [
            function,
        ],
        gamma=0.9,
    )
    assert_raises(RuntimeError, mom, x)


def test_outputs_memory_init_speed_0():
    functions = [
        nn.Sequential(nn.Linear(3, 5), nn.Tanh(), nn.Linear(5, 3))
        for _ in range(36)
    ]
    mom_net = MomentumNet(functions, gamma=0.99)
    mom = Mom(functions, gamma=0.99)
    x = torch.rand(3, requires_grad=True)
    assert torch.allclose(mom_net(x), mom(x), atol=1e-5, rtol=1e-4)
    mom_net_output = (mom_net(x) ** 2).sum()
    mom_output = (mom(x) ** 2).sum()
    assert torch.allclose(
        torch.autograd.grad(mom_net_output, x)[0],
        torch.autograd.grad(mom_output, x)[0],
        atol=1e-5,
        rtol=1e-4,
    )


def test_outputs_memory_init_speed_1():
    functions = [
        nn.Sequential(nn.Linear(3, 5), nn.Tanh(), nn.Linear(5, 3))
        for _ in range(5)
    ]
    init_function = nn.Sequential(nn.Linear(3, 5), nn.Tanh(), nn.Linear(5, 3))
    mom = Mom(functions, gamma=0.99, init_speed=1, init_function=init_function)
    mom_net = MomentumNet(
        functions, gamma=0.99, init_speed=1, init_function=init_function
    )
    x = torch.rand(3, requires_grad=True)
    assert torch.allclose(mom_net(x), mom(x), atol=1e-5, rtol=1e-3)
    x = torch.rand(3, requires_grad=True)
    mom_net_output = (mom_net(x) ** 2 + mom_net(x) ** 3).sum()
    mom_output = (mom(x) ** 2 + mom(x) ** 3).sum()
    params_mom_net = tuple(mom_net.parameters())
    params_mom = tuple(mom.parameters())
    grad_mom_net = torch.autograd.grad(mom_net_output, (x,) + params_mom_net)
    grad_mom = torch.autograd.grad(mom_output, (x,) + params_mom)
    for grad_1, grad_2 in zip(grad_mom_net, grad_mom):
        assert torch.allclose(grad_1, grad_2, atol=1e-5, rtol=1e-3)



# Authors: Michael Sander, Pierre Ablin
# License: MIT

__version__ = '0.8.dev'

from .momentum_net import MomentumNet, MomentumMemory, Mom  # noqa
from .resnet_to_momentumnet import transform # noqa



# Authors: Michael Sander, Pierre Ablin
# License: MIT

"""
Original code from
Maclaurin, Dougal, David Duvenaud, and Ryan Adams.
"Gradient-based hyperparameter optimization through reversible learning."
International conference on machine learning. PMLR, 2015.
"""

import numpy as np
import torch

RADIX_SCALE = 2 ** 52


class TorchExactRep(object):
    def __init__(
        self,
        val,
        from_intrep=False,
        shape=None,
        device=None,
        from_representation=None,
    ):
        if from_representation is not None:
            intrep, store = from_representation
            self.intrep = intrep
            self.aux = BitStore(0, 0, store=store)
        else:
            if device is None:
                device = val.device.type
            if shape is not None:
                self.intrep = torch.zeros(
                    *shape, dtype=torch.long, device=device
                )
            else:
                shape = val.shape
                if from_intrep:
                    self.intrep = val
                else:
                    self.intrep = self.float_to_intrep(val)

            self.aux = BitStore(shape, device)

    def __imul__(self, a):
        self.mul(a)
        return self

    def __iadd__(self, a):
        self.add(a)
        return self

    def __isub__(self, a):
        self.sub(a)
        return self

    def __itruediv__(self, a):
        self.div(a)
        return self

    def add(self, A):
        """Reversible addition of vector or scalar A."""
        self.intrep += self.float_to_intrep(A)
        return self

    def sub(self, A):
        self.add(-A)
        return self

    def rational_mul(self, n, d):
        self.aux.push(self.intrep % d, d)  # Store remainder bits externally
        self.intrep //= d  # Divide by denominator
        self.intrep *= n  # Multiply by numerator
        self.intrep += self.aux.pop(n)  # Pack bits into the remainder

    def mul(self, a):
        n, d = self.float_to_rational(a)
        self.rational_mul(n, d)
        return self

    def div(self, a):
        n, d = self.float_to_rational(a)
        self.rational_mul(d, n)
        return self

    def float_to_rational(self, a):
        d = 2 ** 16 // int(a + 1)
        n = int(a * d + 1)
        return n, d

    def float_to_intrep(self, x):
        if type(x) is torch.Tensor:
            return (x * RADIX_SCALE).long()
        return int(x * RADIX_SCALE)

    def __repr__(self):
        return repr(self.val)

    def n_max_iter(self, beta):
        d, n = self.float_to_rational(beta)
        return int((64 - np.log2(n)) / np.abs(np.log2(n) - np.log2(d)))

    @property
    def val(self):
        return self.intrep.float() / RADIX_SCALE

    def copy(self):
        v = TorchExactRep(self.val)
        v.intrep = torch.clone(self.intrep)
        v.aux.store = torch.clone(self.aux.store)
        return v

    def reset(self):
        self.intrep.fill_(0)
        self.aux.store.fill_(0)


class BitStore(object):
    """
    Efficiently stores information with non-integer number of bits (up to 16).
    """

    def __init__(self, shape, device, store=None):
        # Use an array of Python 'long' ints which conveniently grow
        # as large as necessary. It's about 50X slower though...
        if store is not None:
            self.store = store
        else:
            self.store = torch.zeros(shape, dtype=torch.long).to(device)

    def push(self, N, M):
        """Stores integer N, given that 0 <= N < M"""
        self.store *= M
        self.store += N

    def pop(self, M):
        """Retrieves the last integer stored."""
        N = self.store % M
        self.store //= M
        return N

    def __repr__(self):
        return repr(self.store)



# Authors: Michael Sander, Pierre Ablin
# License: MIT

import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import torchvision
import torchvision.transforms as transforms
import os

from .models import (
    ResNet101,
    mResNet101,
    ResNet18,
    mResNet18,
    mResNet34,
    ResNet34,
    mResNet152,
    ResNet152,
)

n_workers = 10


def train_resnet(
    lr_list,
    model="resnet18",
    mem=False,
    init_speed=0,
    cifar100=False,
    save_adr=None,
    gamma=0.9,
    seed=0,
    save=True,
):
    device = "cuda" if torch.cuda.is_available() else "cpu"
    is_momnet = model.startswith("m")
    # Data
    expe_name = "ckpt_model_%s_seed_%d_gamma_%.2e.pth" % (model, seed, gamma)
    print("==> Preparing data..")
    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(
                (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
            ),
        ]
    )

    transform_test = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize(
                (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
            ),
        ]
    )
    if cifar100:
        Loader = torchvision.datasets.CIFAR100
        root = ".data/CIFAR100"
    else:
        Loader = torchvision.datasets.CIFAR10
        root = ".data/CIFAR10"
    trainset = Loader(
        root=root, train=True, download=True, transform=transform_train
    )
    testset = Loader(
        root=root, train=False, download=True, transform=transform_test
    )
    trainloader = torch.utils.data.DataLoader(
        trainset, batch_size=128, shuffle=True, num_workers=n_workers
    )
    testloader = torch.utils.data.DataLoader(
        testset, batch_size=100, shuffle=False, num_workers=n_workers
    )

    # Model
    print("==> Building model..")
    if model == "mresnet18":
        net = mResNet18
    if model == "resnet18":
        net = ResNet18
    if model == "mresnet34":
        net = mResNet34
    if model == "resnet34":
        net = ResNet34
    if model == "mresnet101":
        net = mResNet101
    if model == "resnet101":
        net = ResNet101
    if model == "mresnet152":
        net = mResNet152
    if model == "resnet152":
        net = ResNet152
    num_classes = 100 if cifar100 else 10
    if not is_momnet:
        net = net(num_classes=num_classes)
    else:
        net = net(
            num_classes=num_classes,
            init_speed=init_speed,
            gamma=gamma,
            mem=mem,
        )
    net = net.to(device)
    if device == "cuda":
        net = torch.nn.DataParallel(net).cuda()
    resume = os.path.isdir("checkpoint_CIFAR10_resnet")
    if resume:
        assert os.path.isdir(
            "checkpoint_CIFAR10_resnet"
        ), "Error: no checkpoint directory found!"
        try:
            checkpoint = torch.load(
                "./checkpoint_CIFAR10_resnet/%s" % expe_name
            )
            net.load_state_dict(checkpoint["net"])
            print("==> Resuming from checkpoint..")
        except OSError:
            pass

    criterion = nn.CrossEntropyLoss().cuda()
    optimizer = optim.SGD(
        net.parameters(), lr=lr_list[0], momentum=0.9, weight_decay=5e-4
    )

    # Training
    def train(net, trainloader, epoch):
        print("\nEpoch: %d" % epoch)
        for param_group in optimizer.param_groups:
            param_group["lr"] = lr_list[epoch]
        net.train()
        train_loss = 0
        correct = 0
        total = 0
        for batch_idx, (inputs, targets) in enumerate(trainloader):
            inputs, targets = inputs.to(device), targets.to(device)
            optimizer.zero_grad()
            outputs = net(inputs)
            loss = criterion(outputs, targets)
            loss.backward()
            optimizer.step()
            train_loss += loss.item()
            _, predicted = outputs.max(1)
            total += targets.size(0)
            correct += predicted.eq(targets).sum().item()
        print(
            "Epoch %d: %.2e, %.2e"
            % (epoch, train_loss / (batch_idx + 1), 100.0 * correct / total)
        )
        return train_loss / (batch_idx + 1), 100.0 * correct / total

    def test(epoch):
        net.eval()
        test_loss = 0
        correct = 0
        total = 0
        with torch.no_grad():
            for batch_idx, (inputs, targets) in enumerate(testloader):
                inputs, targets = inputs.to(device), targets.to(device)
                outputs = net(inputs)
                loss = criterion(outputs, targets)

                test_loss += loss.item()
                _, predicted = outputs.max(1)
                total += targets.size(0)
                correct += predicted.eq(targets).sum().item()
        print(
            "Test  : %.2e, %.2e"
            % (test_loss / (batch_idx + 1), 100.0 * correct / total)
        )
        return test_loss / (batch_idx + 1), 100.0 * correct / total

    train_accs = []
    train_losss = []
    test_losss = []
    test_accs = []

    for epoch in range(len(lr_list)):
        train_loss, train_acc = train(net, trainloader, epoch)
        test_loss, test_acc = test(epoch)
        train_losss.append(train_loss)
        train_accs.append(train_acc)
        test_losss.append(test_loss)
        test_accs.append(test_acc)
        if save:
            if save_adr is not None:
                np.save(
                    save_adr,
                    np.array([train_accs, train_losss, test_accs, test_losss]),
                )
            state = {
                "net": net.state_dict(),
                "acc": test_acc,
                "epoch": epoch,
            }
            if not os.path.isdir("checkpoint_CIFAR10_resnet"):
                os.mkdir("checkpoint_CIFAR10_resnet")
            torch.save(state, "./checkpoint_CIFAR10_resnet/%s" % expe_name)

    return train_accs, train_losss, test_accs, test_losss



# Authors: Michael Sander, Pierre Ablin
# License: MIT

import torch
import torch.nn as nn
from .momentum_net import MomentumNet, Mom


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(
        in_planes, out_planes, kernel_size=1, stride=stride, bias=False
    )


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError(
                "BasicBlock only supports groups=1 and base_width=64"
            )
        if dilation > 1:
            raise NotImplementedError(
                "Dilation > 1 not supported in BasicBlock"
            )
        # Both self.conv1 and self.downsample
        # layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class MBasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(MBasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError(
                "BasicBlock only supports groups=1 and base_width=64"
            )
        if dilation > 1:
            raise NotImplementedError(
                "Dilation > 1 not supported in BasicBlock"
            )
        # Both self.conv1 and self.downsample
        # layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        return out


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride
    # for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places
    # the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for
    # image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5
    # and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion = 4

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample
        # layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)
        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class MBottleneck(nn.Module):
    expansion = 4

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(MBottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample
        # layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)
        return out


class ResNet(nn.Module):
    def __init__(
        self,
        block,
        layers,
        num_classes=1000,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        norm_layer=None,
    ):
        super(ResNet, self).__init__()
        self.large = num_classes == 1000
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer
        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be None"
                " or a 3-element tuple, got {}".format(
                    replace_stride_with_dilation
                )
            )
        self.groups = groups
        self.base_width = width_per_group
        if self.large:
            self.conv1 = nn.Conv2d(
                3,
                self.inplanes,
                kernel_size=7,
                stride=2,
                padding=3,
                bias=False,
            )
        else:
            self.conv1 = nn.Conv2d(
                3, 64, kernel_size=3, stride=1, padding=1, bias=False
            )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(
            block,
            128,
            layers[1],
            stride=2,
            dilate=replace_stride_with_dilation[0],
        )
        self.layer3 = self._make_layer(
            block,
            256,
            layers[2],
            stride=2,
            dilate=replace_stride_with_dilation[1],
        )
        self.layer4 = self._make_layer(
            block,
            512,
            layers[3],
            stride=2,
            dilate=replace_stride_with_dilation[2],
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(
                    m.weight, mode="fan_out", nonlinearity="relu"
                )
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros,
        # and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according
        # to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def _forward_impl(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        if self.large:
            x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.avgpool(x)

        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x

    def forward(self, x):
        return self._forward_impl(x)


class MResNet(nn.Module):
    def __init__(
        self,
        block,
        mblock,
        layers,
        num_classes=1000,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        norm_layer=None,
        gamma=0.9,
        init_speed=0,
        mem=False,
    ):
        super(MResNet, self).__init__()
        self.large = num_classes == 1000
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer
        self.inplanes = 64
        self.dilation = 1
        self.gamma = gamma
        self.mem = mem
        self.init_speed = init_speed
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be"
                " None or a 3-element tuple, got {}".format(
                    replace_stride_with_dilation
                )
            )
        self.groups = groups
        self.base_width = width_per_group
        if self.large:
            self.conv1 = nn.Conv2d(
                3,
                self.inplanes,
                kernel_size=7,
                stride=2,
                padding=3,
                bias=False,
            )
        else:
            self.conv1 = nn.Conv2d(
                3, 64, kernel_size=3, stride=1, padding=1, bias=False
            )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, mblock, 64, layers[0])
        self.layer2 = self._make_layer(
            block,
            mblock,
            128,
            layers[1],
            stride=2,
            dilate=replace_stride_with_dilation[0],
        )
        self.layer3 = self._make_layer(
            block,
            mblock,
            256,
            layers[2],
            stride=2,
            dilate=replace_stride_with_dilation[1],
        )
        self.layer4 = self._make_layer(
            block,
            mblock,
            512,
            layers[3],
            stride=2,
            dilate=replace_stride_with_dilation[2],
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(
                    m.weight, mode="fan_out", nonlinearity="relu"
                )
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros,
        # and each residual block behaves like an identity.
        # This improves the model by
        # 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(
        self, block, mblock, planes, blocks, stride=1, dilate=False
    ):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        if self.mem:
            Maker = Mom
        else:
            Maker = MomentumNet
        for _ in range(1, blocks):

            layers.append(
                mblock(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )
        if self.init_speed != 0:
            init_function = mblock(
                self.inplanes,
                planes,
                groups=self.groups,
                base_width=self.base_width,
                dilation=self.dilation,
                norm_layer=norm_layer,
            )
        else:
            init_function = None
        return nn.Sequential(
            layers[0],
            Maker(
                layers[1:],
                self.gamma,
                init_speed=self.init_speed,
                init_function=init_function,
            ),
        )

    def _forward_impl(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        if self.large:
            x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x

    def forward(self, x):
        return self._forward_impl(x)


def ResNet18(num_classes=1000, bn=True):
    if bn:
        norm_layer = None
    else:
        norm_layer = nn.Identity
    return ResNet(BasicBlock, [2, 2, 2, 2], num_classes, norm_layer=norm_layer)


def mResNet18(num_classes=1000, init_speed=0, gamma=0.9, mem=False, bn=True):
    if bn:
        norm_layer = None
    else:
        norm_layer = nn.Identity
    return MResNet(
        BasicBlock,
        MBasicBlock,
        [2, 2, 2, 2],
        num_classes,
        init_speed=init_speed,
        gamma=gamma,
        mem=mem,
        norm_layer=norm_layer,
    )


def ResNet34(num_classes=1000):
    return ResNet(BasicBlock, [3, 4, 6, 3], num_classes)


def mResNet34(num_classes=1000, init_speed=0, gamma=0.9, mem=False, bn=True):
    if bn:
        norm_layer = None
    else:
        norm_layer = nn.Identity
    return MResNet(
        BasicBlock,
        MBasicBlock,
        [3, 4, 6, 3],
        num_classes,
        init_speed=init_speed,
        gamma=gamma,
        mem=mem,
        norm_layer=norm_layer,
    )


def ResNet101(num_classes=1000):
    return ResNet(Bottleneck, [3, 4, 23, 3], num_classes)


def mResNet101(num_classes=1000, init_speed=0, gamma=0.9, mem=False):
    return MResNet(
        Bottleneck,
        MBottleneck,
        [3, 4, 23, 3],
        num_classes=num_classes,
        init_speed=init_speed,
        gamma=gamma,
        mem=mem,
    )


def ResNet152(num_classes=1000):
    return ResNet(Bottleneck, [3, 8, 36, 3], num_classes)


def mResNet152(num_classes=1000, init_speed=0, gamma=0.9, mem=False):
    return MResNet(
        Bottleneck,
        MBottleneck,
        [3, 8, 36, 3],
        num_classes=num_classes,
        init_speed=init_speed,
        gamma=gamma,
        mem=mem,
    )



# Authors: Michael Sander, Pierre Ablin
# License: MIT

import torch

from momentumnet.exact_rep_pytorch import TorchExactRep


def test_val_torch():
    torch.manual_seed(0)
    x = torch.randn(10)
    x_rep = TorchExactRep(x)
    x_init_val = x_rep.val.clone()
    assert ((x - x_rep.val).abs() < 1e-7).byte().all()
    x_rep *= 3.0
    x *= 3.0
    assert ((x - x_rep.val).abs() < 1e-3).byte().all()
    x_rep += 2
    x += 2
    assert ((x - x_rep.val).abs() < 1e-3).byte().all()
    x_rep -= 2
    x -= 2
    assert ((x - x_rep.val).abs() < 1e-3).byte().all()
    x_rep /= 3
    x /= 3
    assert ((x - x_rep.val).abs() < 1e-3).byte().all()
    assert (x_rep.val == x_init_val).byte().all()



# Authors: Michael Sander, Pierre Ablin
# License: MIT
import torch
import torch.nn as nn
from torch.nn.parameter import Parameter
from .exact_rep_pytorch import TorchExactRep


class MomentumNet(nn.Module):
    def __init__(
        self,
        functions,
        gamma,
        n_iters=1,
        learn_gamma=False,
        init_speed=0,
        init_function=None,
    ):
        super(MomentumNet, self).__init__()
        if gamma < 0 or gamma > 1:
            raise Exception("gamma has to be between 0 and 1")
        self.n_functions = len(functions)
        for i, function in enumerate(functions):
            self.add_module(str(i), function)
        self.learn_gamma = learn_gamma
        if learn_gamma:
            self.gamma = Parameter(torch.tensor(gamma))
        else:
            self.gamma = gamma
        self.n_iters = n_iters
        self.init_speed = init_speed
        if init_function is not None:
            self.add_module("init", init_function)

    def forward(self, x, n_iters=None, ts=1):
        if n_iters is None:
            n_iters = self.n_iters
        if self.init_speed == 0:
            v = torch.zeros_like(x)
        else:
            v = self.init_function(x)
        gamma = self.gamma
        for i in range(n_iters):
            for function in self.functions:
                v = gamma * v + function(x) * ts * (1 - gamma)
                x = x + v * ts
        return x

    @property
    def functions(self):
        return [self._modules[str(i)] for i in range(self.n_functions)]

    @property
    def init_function(self):
        return self._modules["init"]


class MomentumMemory(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, v, gamma, functions, init_function, *params):
        ctx.functions = functions
        ctx.gamma = gamma
        ctx.init_function = init_function
        v = TorchExactRep(v)
        with torch.no_grad():
            for function in functions:
                v *= gamma
                v += (1 - gamma) * function(x)
                x = x + v.val
        ctx.save_for_backward(x, v.intrep, v.aux.store)
        return x

    @staticmethod
    def backward(ctx, grad_output):
        functions = ctx.functions
        gamma = ctx.gamma
        init_function = ctx.init_function
        x, v_intrep, v_store = ctx.saved_tensors
        v = TorchExactRep(0, from_representation=(v_intrep, v_store))
        grad_x = grad_output
        grad_v = torch.zeros_like(grad_x)
        grad_params = []
        with torch.set_grad_enabled(True):
            for function in functions[::-1]:
                x = x.detach().requires_grad_(False)
                x = x - v.val
                x = x.detach().requires_grad_(True)
                f_eval = function(x)
                grad_combi = grad_x + grad_v
                vjps = torch.autograd.grad(
                    f_eval, (x,) + tuple(function.parameters()), grad_combi
                )
                v += -(1 - gamma) * f_eval
                v /= gamma
                grad_params.append([(1 - gamma) * vjp for vjp in vjps[1:]])
                grad_x = grad_x + (1 - gamma) * vjps[0]
                grad_v = gamma * grad_combi
            if ctx.init_function is not None:
                x = x.detach().requires_grad_(True)
                f_eval = init_function(x)
                params = tuple(init_function.parameters())
                vjps = torch.autograd.grad(
                    f_eval, params, torch.zeros_like(grad_x)
                )
                grad_params.append([vjp for vjp in vjps])
            else:
                pass
        flat_params = []
        for param in grad_params[::-1]:
            flat_params += param
        return (grad_x, grad_v, None, None, None, *flat_params)


class Mom(nn.Module):
    def __init__(self, functions, gamma, init_speed=0, init_function=None):
        super(Mom, self).__init__()
        if gamma < 0 or gamma > 1:
            raise Exception("gamma has to be between 0 and 1")
        for i, function in enumerate(functions):
            self.add_module(str(i), function)
        self.gamma = gamma
        self.init_speed = init_speed
        self.n_functions = len(functions)
        for i, function in enumerate(functions):
            self.add_module(str(i), function)
        self.v = None
        if init_function is not None:
            self.add_module("init", init_function)

    def forward(self, x):
        if self.init_speed == 0:
            init_function = None
            v = torch.zeros_like(x)
            params = []
        else:
            v = self.init_function(x)
            params = list(self.init_function.parameters())
            init_function = self.init_function
        functions = self.functions
        for function in functions:
            params += list(function.parameters())
        output = MomentumMemory.apply(
            x, v, self.gamma, functions, init_function, *params
        )
        self.v = v
        return output

    def inverse(self, x):
        v = self.v
        gamma = self.gamma
        with torch.no_grad():
            for function in self.functions[::-1]:
                x = x - v.val
                v -= (1 - gamma) * function(x)
                v /= gamma
        return x

    @property
    def functions(self):
        return [self._modules[str(i)] for i in range(self.n_functions)]

    @property
    def init_function(self):
        return self._modules["init"]



#! /usr/bin/env python

import os
from setuptools import setup

descr = """Momentum Residual Neural Networks"""

version = None
with open(os.path.join("momentumnet", "__init__.py"), "r") as fid:
    for line in (line.strip() for line in fid):
        if line.startswith("__version__"):
            version = line.split("=")[1].strip().strip("'")
            break
if version is None:
    raise RuntimeError("Could not determine version")


DISTNAME = "momentumnet"
DESCRIPTION = descr
MAINTAINER = "Michael E. Sander"
MAINTAINER_EMAIL = "michael.sander1295@yahoo.fr"
LICENSE = "MIT"
DOWNLOAD_URL = "https://github.com/michaelsdr/momentumnet.git"
VERSION = version
URL = "https://github.com/michaelsdr/momentumnet"


def get_requirements():
    """Return the requirements of the projects in requirements.txt"""
    with open("requirements.txt") as f:
        requirements = [r.strip() for r in f.readlines()]
    return [r for r in requirements if r != ""]


def package_tree(pkgroot):
    """Get the submodule list."""
    # Adapted from VisPy
    path = os.path.dirname(__file__)
    subdirs = [
        os.path.relpath(i[0], path).replace(os.path.sep, ".")
        for i in os.walk(os.path.join(path, pkgroot))
        if "__init__.py" in i[2]
    ]
    return sorted(subdirs)


if __name__ == "__main__":
    setup(
        name=DISTNAME,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        license=LICENSE,
        version=VERSION,
        url=URL,
        download_url=DOWNLOAD_URL,
        long_description=open('README.rst').read(),
        long_description_content_type='text/x-rst',
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "License :: OSI Approved",
            "Programming Language :: Python",
            "Topic :: Software Development",
            "Topic :: Scientific/Engineering",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Operating System :: MacOS",
            "Programming Language :: Python :: 3",
        ],
        platforms="any",
        python_requires=">=3.8",
        packages=package_tree("momentumnet"),
        install_requires=get_requirements(),
    )
